"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CommandBehavior = exports.AwsStub = void 0;
const sinon_1 = require("sinon");
const mockClient_1 = require("./mockClient");
/**
 * Wrapper on the mocked `Client#send()` method,
 * allowing to configure its behavior.
 *
 * Without any configuration, `Client#send()` invocation returns `undefined`.
 *
 * To define resulting variable type easily, use {@link AwsClientStub}.
 */
class AwsStub {
    constructor(client, send) {
        this.client = client;
        this.send = send;
    }
    /** Returns the class name of the underlying mocked client class */
    clientName() {
        return this.client.constructor.name;
    }
    /**
     * Resets stub. It will replace the stub with a new one, with clean history and behavior.
     */
    reset() {
        /* sinon.stub.reset() does not remove the fakes which in some conditions can break subsequent stubs,
         * so instead of calling send.reset(), we recreate the stub.
         * See: https://github.com/sinonjs/sinon/issues/1572
         * We are only affected by the broken reset() behavior of this bug, since we always use matchers.
         */
        const newStub = (0, mockClient_1.mockClient)(this.client);
        this.send = newStub.send;
        return this;
    }
    /** Resets stub's calls history. */
    resetHistory() {
        this.send.resetHistory();
        return this;
    }
    /** Replaces stub with original `Client#send()` method. */
    restore() {
        this.send.restore();
    }
    /**
     * Returns recorded calls to the stub.
     * Clear history with {@link resetHistory} or {@link reset}.
     */
    calls() {
        return this.send.getCalls();
    }
    /**
     * Returns n-th recorded call to the stub.
     */
    call(n) {
        return this.send.getCall(n);
    }
    /**
     * Returns recorded calls of given Command only.
     * @param commandType Command type to match
     * @param input Command payload to match
     * @param strict Should the payload match strictly (default false, will match if all defined payload properties match)
     */
    commandCalls(commandType, input, strict) {
        return this.send.getCalls()
            .filter((call) => {
            const isProperType = call.args[0] instanceof commandType;
            const inputMatches = this.createInputMatcher(input, strict).test(call.args[0]);
            return isProperType && inputMatches;
        });
    }
    onAnyCommand(input, strict = false) {
        const cmdStub = this.send.withArgs(this.createInputMatcher(input, strict));
        return new CommandBehavior(this, cmdStub);
    }
    on(command, input, strict = false) {
        const matcher = sinon_1.match.instanceOf(command).and(this.createInputMatcher(input, strict));
        const cmdStub = this.send.withArgs(matcher);
        return new CommandBehavior(this, cmdStub);
    }
    createInputMatcher(input, strict = false) {
        return input !== undefined ?
            sinon_1.match.has('input', strict ? input : (0, sinon_1.match)(input))
            : sinon_1.match.any;
    }
    resolves(response) {
        return this.onAnyCommand().resolves(response);
    }
    resolvesOnce(response) {
        return this.onAnyCommand().resolvesOnce(response);
    }
    rejects(error) {
        return this.onAnyCommand().rejects(error);
    }
    rejectsOnce(error) {
        return this.onAnyCommand().rejectsOnce(error);
    }
    callsFake(fn) {
        return this.onAnyCommand().callsFake(fn);
    }
    callsFakeOnce(fn) {
        return this.onAnyCommand().callsFakeOnce(fn);
    }
}
exports.AwsStub = AwsStub;
class CommandBehavior {
    constructor(clientStub, send) {
        this.clientStub = clientStub;
        this.send = send;
        /**
         * Counter to simulate chainable `resolvesOnce()` and similar `*Once()` methods with Sinon `Stub#onCall()`.
         * The counter is increased with every `*Once()` method call.
         */
        this.nextChainableCallNumber = 0;
        /**
         * Function to get the current Client object from inside the `callsFake()` callback.
         * Since this is called from the callback when the mock function is executed,
         * the current Client is the last on the Sinon `Stub#thisValues` list.
         */
        this.getClient = () => this.send.thisValues[this.send.thisValues.length - 1];
    }
    onAnyCommand(input, strict) {
        return this.clientStub.onAnyCommand(input, strict);
    }
    on(command, input, strict = false) {
        return this.clientStub.on(command, input, strict);
    }
    resolves(response) {
        this.send.resolves(response);
        return this.clientStub;
    }
    resolvesOnce(response) {
        this.send = this.send.onCall(this.nextChainableCallNumber++).resolves(response);
        return this;
    }
    rejects(error) {
        this.send.rejects(CommandBehavior.normalizeError(error));
        return this.clientStub;
    }
    rejectsOnce(error) {
        this.send.onCall(this.nextChainableCallNumber++).rejects(CommandBehavior.normalizeError(error));
        return this;
    }
    static normalizeError(error) {
        if (typeof error === 'string') {
            return new Error(error);
        }
        if (!(error instanceof Error)) {
            return Object.assign(new Error(), error);
        }
        return error;
    }
    callsFake(fn) {
        this.send.callsFake(cmd => this.fakeFnWrapper(cmd, fn));
        return this.clientStub;
    }
    callsFakeOnce(fn) {
        this.send.onCall(this.nextChainableCallNumber++).callsFake(cmd => this.fakeFnWrapper(cmd, fn));
        return this;
    }
    fakeFnWrapper(cmd, fn) {
        try {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            return fn(cmd.input, this.getClient);
        }
        catch (err) {
            return Promise.reject(CommandBehavior.normalizeError(err));
        }
    }
}
exports.CommandBehavior = CommandBehavior;
//# sourceMappingURL=awsClientStub.js.map